/**PrologFile*************************************************************

   Filename     [ $Source: /home/schumann/CVS/PN/synth/schemas/gsl/gsl-maximization.pl,v $ ]

   Synopsis     [ Function maximization using GSL-routines ]

   Author       [ 
   		  Bernd Fischer           <fisch@email.arc.nasa.gov>
		]

   Revision     [ $Id: gsl-maximization.pl,v 1.2 2003/02/14 04:40:04 fisch Exp $ ]

   Description  [ 

                ]

   Known Bugs   [ None. ]

   Modification [

	$Log: gsl-maximization.pl,v $
	Revision 1.2  2003/02/14 04:40:04  fisch
	Minor bugfixes for cse-calls.
	
	Revision 1.1  2003/01/16 18:28:36  fisch
	Initial revision.
	

		]

 *************************************************************************/

:- ensure_loaded([
	'synth/schemas/gsl/gsl-abstraction.pl'
   ]).

%%% this is a hack - it should go somewhere else...
%%% make C's address-operator visible to the expression pretty-printer
:- op(1200, fx, '&').
:- op(1200, fx, '* ').


/*************************************************************************
 *
 * One-dimensional optimization
 *
 *************************************************************************
 *
 * The GSL provides two one-dimensional optimization algorithms, Brent's
 * algorithm, and Golden Section Search. However, both algorithms use the
 * same overall control structure, so the schema constructs both versions
 * as successive solutions.
 *
 *************************************************************************/

synth_formula(Theta, Formula, Constraint, Program) :-

	%
	% Iterate over the two methods
	%
	member([Method, GSLName], 
	       [['Brent-Method',          'gsl_min_fminimizer_brent'],
	        ['Golden Section Search', 'gsl_min_fminimizer_goldensection']
	       ]
	      ),

        synth_trace_schema_enter(
        	'synth/schemas/gsl/gsl-maximization.pl',
        	['GSL-routine', GSLName, '(', Method, ')'],
                synth_formula(Theta, Formula, Constraint, Program)
        ),


        /*****
         *
         * Guards
         *
         *****/

	%
	% Check whether GSL is enabled
	%
	flag(lib_gsl, true, true),

	%
	% Check whether numeric solutions are permissible
	%
	\+ flag(synth_force_symbolic, true, true),

	%
	% Check single-variable case
        %
	Theta = [Var],
	ground(Var),

	%
	% Find (enclosing) non-trivial range for Var and initialize bracket
	%
	% NOTE: The bracket variables A and B are still uninstantiated.
	%
	synth_init_range(_, Var, A, B, Init0),


        /*****
         *
         * Body
         *
         *****/

        assert_trace(trace_schema_inout,
                     'synth/schemas/gsl/gsl-maximization.pl',
                     ['Applying GSL-routine (', Method, ')']
                    ),

	%
	% Generate GSL-function
	%
	% NOTE: The GSL-functions minimize, so we have to negate the
	%	formula.
	%
	simplify(Assumptions, -Formula, NFormula),
	gsl_function(NFormula, Assumptions, Var, Decls, Init, Name),

        %
        % Generate dummy declaration for parameters
        %
        model_gensym(dummy, Dummy),
        D_decl = native(Dummy,
                        'void *',
                        [comment('dummy parameters required by GSL')]
                       ),

        %
        % Generate bracket names
        %
        model_gensym(bracket_a, A),
        model_gensym(bracket_b, B),
        model_gensym(bracket_c, C),
        A_decl = scalar(A, double, [comment(['Bracket - "left" border'])]),
        B_decl = scalar(B, double, [comment(['Bracket - central point (with maximal value)'])]),
        C_decl = scalar(C, double, [comment(['Bracket - "right" border'])]),

	%
	% Bracketing
	%
	FCall =.. [Name, Var, Dummy],
        synth_bracket_max(FCall, Var, A, B, C, Bracketing),

	%
	% Swap the found brackt borders such that A =< B
	%
	Swap = if(A > B,
		  block(local([scalar(PV, double, [])]),
			series([assign(PV, A, []),
			        assign(A, B, []),
			        assign(B, A, [])
			       ],
			       []
			      ),
			[]
		       ),
		  skip,
		  [comment(['Order bracket borders such that ', A, 
			    ' is the left border'])]
		 ),

	%
	% Declarations for the search loop
	%
	gsl_status_decl(Status, Status_decl),
	gsl_count_decl(Count, Count_decl),
	gsl_countmax_decl(CountMax, Max_decl),
	gsl_mtype_decl('gsl_min_fminimizer_type', MType, MType_decl),
	gsl_mstate_decl('gsl_min_fminimizer', MState, MState_decl),
	gsl_mfun_decl('gsl_function', MFun, MFun_decl),

	%
	% Setup of the search loop
	%
	atom_concat(MFun, '.function', MFun_function),
	atom_concat(MFun, '.params',   MFun_params),
	FSetup = series([assign(MFun_function, '&'(Name), []), 
			 assign(MFun_params, 0, [])
			], 
			[comment('Setup the GSL target function')]
		       ),
	OSetup = series([assign(MType, GSLName, []), 
			 assign(MState, gsl_min_fminimizer_alloc(MType), []),
			 call(gsl_min_fminimizer_set(MState, '&'(MFun), B, A, C), [])
			], 
			[comment('Setup the GSL optimizer')]
		       ),
	Setup  = series([assign(Count, 0, []), 
			 assign(CountMax, 100, []), 
			 FSetup, 
			 OSetup
			], 
			[comment('Setup the optimization loop'),
			 label(_Setup_label)
			]
		       ),
	Free = call(gsl_min_fminimizer_free(MState), 
		    [comment('Free the GSL optimizer')]
		   ),

	%
	% Search loop
	%
	Check = if(Status = 'GSL_SUCCESS',
		   series([assign_plus(Count, 1, []),
			   series([assign(A, gsl_min_fminimizer_x_lower(MState), []),
			           assign(C, gsl_min_fminimizer_x_upper(MState), [])
			          ],
			          [comment('Get current error interval')]
			         ),
			   assign(Status, 
				  gsl_min_test_interval(A, C, tolerance, 0), 
				  [comment('Check error interval (no relative error)')]
				 )
			  ],
			  []
			 ),
		   ab_error(gsl_error(Status), []),
		   []
		  ),
	Loop  = repeat(or(Status = 'GSL_SUCCESS', Count >= CountMax),
		       series([assign(Status, gsl_min_fminimizer_iterate(MState), []),
			       Check
			      ], 
			      []
			     ),
		       [comment('Optimization loop')]
		      ),

        %
        % Construct explanation
        %
        explain_get_attributes(label, Bracketing, label(Bracketing_label)),
        XP = [
          'The function', expr(Formula), 'is then maximized w.r.t. the ',
          'goal variable ', Var, ' using the GSL-routine \"', GSLName,
	  '\" (i.e., ', Method, '). This routine requires an initial ',
	  'bracket for the maximum ', labelref(Bracketing_label),
          ' before it starts to search for the maximum itself ',
          labelref(Setup_label), '.'
        ],

	%
	% Compose program
	%
	Program = block(local([A_decl, B_decl, C_decl, D_decl,
			       Status_decl, Count_decl, Max_decl,
			       MType_decl, MState_decl, MFun_decl
			       | Decls]
			     ),
			series([Init, Init0, Bracketing, Swap, Setup, Loop, Free],[]),
			[comment(XP)]
		       ),

        assert_trace(trace_schema_inout,
                     'synth/schemas/gsl/gsl-maximization.pl',
        	     ['Done GSL-routine', GSLName, '(', Method, ')']
                    ).

/*************************************************************************
 *
 * Multi-dimensional optimization / conjugate gradient methods
 *
 *************************************************************************
 *
 * The GSL provides three multi-dimensional conjugate gradient algorithms,
 * the Fletcher-Reeves method, the Polak-Ribiere method, and the Broyden-
 * Fletcher-Goldfarb-Shanno method. However, all three algorithms use the
 * same overall control structure, so the schema constructs all versions
 * as successive solutions.
 *
 *************************************************************************/

synth_formula(Theta, Formula, Constraint, Program) :-

	%
	% Iterate over the three methods
	%
	member([Method, GSLName], 
	       [['Fletcher-Reeves Conjugate Gradient Method',
		 'gsl_multimin_fdfminimizer_fr'],
	        ['Polak-Ribiere Conjugate Gradient Method',
		 'gsl_multimin_fdfminimizer_pr'],
	        ['Broyden-Fletcher-Goldfarb-Shanno Conjugate Gradient Method',
		 'gsl_multimin_fdfminimizer_bfgs']
	       ]
	      ),

        synth_trace_schema_enter(
        	'synth/schemas/gsl/gsl-maximization.pl',
        	['GSL-routine', GSLName, '(', Method, ')'],
                synth_formula(Theta, Formula, Constraint, Program)
        ),


        /*****
         *
         * Guards
         *
         *****/

	%
	% Check whether GSL is enabled
	%
	flag(lib_gsl, true, true),

	%
	% Check whether numeric solutions are permissible
	%
	\+ flag(synth_force_symbolic, true, true),

	%
	% Check multi-variable case
        %
	(  \+ ground(Theta)
	 ; 
	   Theta = [_, _ | _]
	),


        /*****
         *
         * Body
         *
         *****/

        assert_trace(trace_schema_inout,
                     'synth/schemas/gsl/gsl-maximization.pl',
                     ['Applying GSL-routine (', Method, ')']
                    ),

	%
	% Compute offset vector and size
	%
	apack_offset_vector(Theta, Offsets),
	append(_, [[_, LOffset, LSize]], Offsets),
	simplify(Assumptions, LOffset + LSize, N),

	%
	% Generate GSL-functions
	%
	% NOTE: The GSL-functions minimize, so we have to negate the
	%	formula.
	%
	simplify(Assumptions, -Formula, NFormula),
	gsl_fdf(NFormula, Assumptions, Offsets, Decls, Init, FName, DFName, FDFName),


	%
	% Declarations for the search loop
	%
	gsl_status_decl(Status, Status_decl),
	gsl_count_decl(Count, Count_decl),
	gsl_countmax_decl(CountMax, Max_decl),
	gsl_mtype_decl('gsl_multimin_fdfminimizer_type', MType, MType_decl),
	gsl_mstate_decl('gsl_multimin_fdfminimizer', MState, MState_decl),
	gsl_mfun_decl('gsl_multimin_function_fdf', MFun, MFun_decl),

	%
	% Declaration for the argument vector
	%
	model_gensym(gsl_args, X),
	X_decl = native(X, 'gsl_vector *', 
			[comment('Argument vector for GSL functions')]
		       ),

	%
	% Initialize the argument vector
	%
	maplist(synth_init_start, Theta, ITheta),
	apack_update(call, 
		     gsl_vector_set(X),
		     [comment('Initialize argument vector')], 
		     Offsets, 
		     ITheta, 
		     XInit),

	%
	% Setup of the search loop
	%
	atom_concat(MFun, '.f',      MFun_f),
	atom_concat(MFun, '.df',     MFun_df),
	atom_concat(MFun, '.fdf',    MFun_fdf),
	atom_concat(MFun, '.n',      MFun_n),
	atom_concat(MFun, '.params', MFun_params),
	FSetup = series([assign(MFun_f,     '&'(FName), []), 
			 assign(MFun_df,    '&'(DFName), []),
			 assign(MFun_fdf,   '&'(FDFName), []),
			 assign(MFun_n,      N, []),
			 assign(MFun_params, 0, [])
			], 
			[comment('Setup the GSL target function')]
		       ),
	XSetup = series([assign(X, gsl_vector_alloc(N), []),
		         XInit
		        ],
		        [comment('Setup the argument vector')]
		       ),
	OSetup = series([assign(MType, GSLName, []), 
			 assign(MState, gsl_min_fminimizer_alloc(MType, N), []),
			 call(gsl_multimin_fdfminimizer_set(MState, '&'(MFun), X, 0.01, tolerance/10), [])
			], 
			[comment('Setup the GSL optimizer')]
		       ),
	Setup  = series([assign(Count, 0, []), 
			 assign(CountMax, 100, []), 
			 FSetup, 
			 XSetup, 
			 OSetup
			], 
			[comment('Setup the optimization loop'),
			 label(Setup_label)
			]
		       ),
	Free = series([call(gsl_multimin_fdfminimizer_free(MState), []),
		       call(gsl_vector_free(X), [])
		      ],
		      [comment('Free the GSL optimizer and argument vector')]
		     ),

	%
	% Search loop
	%
	atom_concat(MState, '->gradient', Gradient),
	Check = if(Status = 'GSL_SUCCESS',
		   series([assign_plus(Count, 1, []),
			   assign(Status, 
				  gsl_min_test_gradient(Gradient, tolerance), 
				  []
				 )
			  ],
			  []
			 ),
		   ab_error(gsl_error(Status), []),
		   []
		  ),
	Loop  = repeat(or(Status = 'GSL_SUCCESS', Count >= CountMax),
		       series([assign(Status, gsl_multimin_fdfminimizer_iterate(MState), []),
			       Check
			      ], 
			      []
			     ),
		       [comment('Optimization loop')]
		      ),

        %
        % Construct explanation
        %
	lex_enumerate_vars(Theta, XP_theta),
        XP = [
          'The function', expr(Formula), 'is then maximized w.r.t. the ',
          'goal ', XP_theta, ' using the GSL-routine \"', GSLName,
	  '\" (i.e., the ', Method, ').'
        ],

	%
	% Compose the program
	%
	Program = block(local([Status_decl, Count_decl, Max_decl, X_decl,
			       MType_decl, MState_decl, MFun_decl
			       | Decls]
			     ),
			series([Init, Setup, Loop, Free], []),
			[comment(XP)]
		       ),

        assert_trace(trace_schema_inout,
                     'synth/schemas/gsl/gsl-maximization.pl',
        	     ['Done GSL-routine', GSLName, '(', Method, ')']
                    ).



/*************************************************************************
 *
 * Auxiliary predicates
 *
 *************************************************************************/


/*****
 *
 * Generate Declarations
 *
 * The following set of predicates generates (one by one) the declartions
 * required for the optimization routines.
 *
 *****/

gsl_status_decl(Var, Decl) :-
	model_gensym(status, Var),
	Decl = scalar(Var, int, [comment('Status of GSL optimization step')]).

gsl_count_decl(Var, Decl) :-
	model_gensym(count, Var),
	Decl = scalar(Var, int, [comment('Number of elapsed optimization steps')]).

gsl_countmax_decl(Var, Decl) :-
	model_gensym(countmax, Var),
	%%% this should be a constant but blocks don't support local constants
	Decl = scalar(Var, int, [comment('Maximum number of optimization steps')]).

gsl_mtype_decl(Type, Var, Decl) :-
	model_gensym(mtype, Var),
	%
	% NOTE: This needs to be a constant but blocks don't support local 
	%	constants, so it is encoded into the C-type.
	%
	concat_atom(['const ', Type, ' *'], CType),
	Decl = native(Var, CType, [comment('GSL optimizer type')]).

gsl_mstate_decl(Type, Var, Decl) :-
	model_gensym(mstate, Var),
	atom_concat(Type, ' *', CType),
	Decl = native(Var, CType, [comment('GSL optimizer state')]).

gsl_mfun_decl(Type, Var, Decl) :-
	model_gensym('F', Var),
	Decl = native(Var, Type, [comment('GSL target function')]).

