/**PrologFile*************************************************************

   Filename     [ $Source: /home/schumann/CVS/PN/synth/schemas/numeric/simplex.pl,v $ ]

   Synopsis     [ Nelder-Mead simplex routine ]


   Author       [ 
   		  Bernd Fischer           <fisch@ptolemy.arc.nasa.gov>
		]

   Revision     [ $Id: simplex.pl,v 1.17 2003/01/25 01:04:18 fisch Exp $ ]

   Description  [ 

                ]

   Known Bugs   [ ]

   Modification [

	$Log: simplex.pl,v $
	Revision 1.17  2003/01/25 01:04:18  fisch
	Replaced simplex-packing/unpacking by apack-routines.
	
	Revision 1.16  2002/12/08 00:10:32  fisch
	Changed/completed explanations.
	
	Revision 1.15  2002/11/16 02:12:18  fisch
	Fixed bug in constructing unpacking of simplex values.
	
	Revision 1.14  2002/11/15 19:32:45  fisch
	Array packing/unpacking now done by apack_*-predicates.
	
	Revision 1.13  2002/11/01 23:36:16  fisch
	Changed generated initialization and explanations.
	
	Revision 1.12  2002/10/24 21:15:44  fisch
	Added control flag to schema.

	Revision 1.11  2002/10/16 21:45:15  fisch
	Modified explanations.

	Revision 1.10  2002/10/16 21:42:04  fisch
	Fixed convergence loop by adding worst and second-worst point to
	convergence variables.

	Revision 1.9  2002/10/03 16:21:41  fisch
	Added common subexpression elimination.

	[...]

	Revision 1.1  2001/10/13 01:45:36  wray
	Moved out of synth.pl

		]

 *************************************************************************/

:- ensure_loaded([
	'synth/apack.pl',
	'synth/synth.pl',
	'synth/schemas/numeric/support.pl',
	'optimize/cse.pl'
   ]).

/*************************************************************************
 *
 * synth_formula/4 - Numerical solution (Nelder-Mead Simplex Method)
 *
 *
 *
 * References:
 * J. A. Nelder and R. Mead.
 *
 * William H. Press, Saul A. Teukolsky, William T. Vetterling and 
 * Brian P. Flannery. Numerical Recipes in C (2nd Edition), Ch. 10.4. 
 * Cambridge University Press. Cambridge, UK, 1992.
 *
 *************************************************************************/

synth_formula(Theta, Formula, Constraint, Program) :-

	assert_trace(trace_schema_head, 'synth/schemas/numeric/simplex.pl',
		['Trying Nelder-Mead simplex method:',
		  synth_formula(Theta, Formula, Constraint, Program)
		]),

	%
	% Array base & offset, 0-based arrays
	%
	% NOTE: Set Ar_base = 1, Ar_offset = 0 for 1-based arrays
	%
	% NOTE: Hardcode, or get from environment
	%
	Ar_base   = 0,
	Ar_offset = 1,

	/*****
	 *
	 * Guards
	 *
	 *****/

        %
        % Check whether numeric solutions are permissible
        %
        \+ flag(synth_force_symbolic, true, true),

	%
	% Optimize for more than a single variable / vector instance
	%   (--> use simpler methods otherwise)
	%
	(Theta = [_, _ | _] ; \+ ground(Theta)),

	%
	% Each variable in Theta admits initial values and step sizes
	%
	list_unzip(synth_init_start_step, Theta, Init_start, Init_step),


	/*****
	 *
	 * Body
	 *
	 *****/

	assert_trace(trace_schema_inout,
		     'synth/schemas/numeric/simplex.pl', 
		     'Applying Nelder-Mead simplex method'),

	%
	% Find subexpressions unrelated to the goal variables and
	% let-abstract them out; this also restructures the goal
	% function.
	%
	cse_unrelated(Formula, Theta, Constraint, SFormula, CDecls, CInit, XPSexp), 

	%
	% Compute the offset-vector for packing into the simplex-array;
	% also calculate the dimension of the search space
	%
	apack_offset_vector(Theta, Offset_vector),
	append(_, [[_, LOffset, LSize]], Offset_vector),
	simplify(LOffset + LSize, Size),

        /*
         * declarations for simplex,  point-vector, center, and trial point
         *
         */
	model_gensym(simplex, Simplex),
	simplify(Size+1-Ar_offset, Size1),
	simplify(Size-Ar_offset, Size0),
	Simplex_dim  = [dim(Ar_base, Size1), dim(Ar_base, Size0)], 
	Simplex_decl = matrix(Simplex, double, 
			      Simplex_dim,
			      [comment(['simplex structure used in Nelder-Mead simplex method ', 
		                        labelref(Simplex_label), pp_nl,
					'(', Size, '+1) points in the ', Size, 
					'-dimensional space'
				       ])
			      ]
			     ),

	model_gensym(s_values, Values),
	Values_dim  = dim(Ar_base, Size1),
	Values_decl = vector(Values, double, 
			     [Values_dim],
			     [comment(['function values of the simplex ',
				       labelref(Simplex_label)
				      ])
			     ]
			    ),

	model_gensym(s_center, Center),
	Center_dim  = dim(Ar_base, Size0),
	Center_decl = vector(Center, double, 
			     [Center_dim],
			     [comment(['center of gravity of the simplex ',
				       labelref(Simplex_label), pp_nl,
					'single point in the ', Size, 
					'-dimensional space'
				      ])
			     ]
			    ),

	model_gensym(s_try, Try),
	Try_decl = vector(Try, double, 
			  [dim(Ar_base, Size0)],
			  [comment(['trial point used in Nelder-Mead simplex method ', 
				    labelref(Simplex_label), pp_nl,
				    'single point in the ', Size, 
				    '-dimensional space'
				   ])]),

	model_gensym(s_tryval, Try_val),
	Try_val_decl = scalar(Try_val, double, 
			      [comment(['function value at trial point ',
				        labelref(Simplex_label)
				       ])]),

	model_gensym(s_trysave, Save_val),
	Save_val_decl = scalar(Save_val, double, []),

        /*
         * initialize simplex
         *
         */
	copy_term(Offset_vector, Offset_vector_copy),
	synth_simplex_init(Simplex, Size, Simplex_dim, 
			   Offset_vector, Init_start, Init_step,
			   Init_simplex),
        /*
         * evaluate function at the simplex
         *
         */
	pv_fresh_existential(I),
	index_make([I, Values_dim], Index_i),
	Value_i   =.. [Values, I],
	Simplex_i =.. [Simplex, I],
	apack_substitute(Simplex_i, Offset_vector, SFormula, SFormula_i),
	Init_values = for([Index_i],
			  assign(Value_i, SFormula_i, []),
			  [comment('Evaluate expression at each simplex-point')]
			 ),

        /*
         * calculate the center of gravity
         *
         */
	pv_fresh_existential(J1),
	pv_fresh_existential(J2),
	index_make([J1, dim(Ar_base, Size0)], Index_j1),
	index_make([J2, dim(Ar_base, Size1)], Index_j2),
	Center_j   =.. [Center, J1],
	Simplex_ij =.. [Simplex, J2, J1],
	Centroid = for([Index_j1],
		       assign(Center_j, 
			      sum([Index_j2], Simplex_ij), 
			      []),
		       [comment('Calculate the center of gravity in the simplex')]
		       ),

        /*
         * find the face of the simplex 
         *
         */
	model_gensym(s_best,   Best),
	model_gensym(s_sworst, Worst2),
	model_gensym(s_worst,  Worst1),
	Best_decl   = scalar(Best, int, 
			     [comment(['simplex-face: best point ',
				       labelref(Simplex_label)])]),
	Worst2_decl = scalar(Worst2, int, 
			     [comment(['simplex-face: second-worst point ',
				       labelref(Simplex_label)])]),
	Worst1_decl = scalar(Worst1, int, 
			     [comment(['simplex-face: worst point ',
				       labelref(Simplex_label)])]),
	Values0  =.. [Values, Ar_base],
	Values1  =.. [Values, Ar_base+1],
	ValuesK  =.. [Values, K],
	ValuesB  =.. [Values, Best],
	ValuesW2 =.. [Values, Worst2],
	ValuesW1 =.. [Values, Worst1],
	Sweep_loop = for([idx(K, Ar_base+2, Size1)],
			 if(ValuesK > ValuesB, 
			    assign(Best, K, []), 
			    if(ValuesK =< ValuesW1, 
			       series([assign(Worst2, Worst1, []),
				       assign(Worst1, K, [])
				      ], []), 
			       if(ValuesK =< ValuesW2,
				  assign(Worst2, K, []),
				  skip,
				  []),
			       []
			      ),
			    []
		           ),
			 []
			),
	Sweep = series([if(Values0 < Values1,
			series([assign(Best,   Ar_base+1, []),
			  	assign(Worst1, Ar_base, []),
				assign(Worst2, Ar_base+1, [])
			       ], []),
			series([assign(Best,   Ar_base, []),
			        assign(Worst1, Ar_base+1, []),
				assign(Worst2, Ar_base, [])
			       ], []),
			[]),
			Sweep_loop
		       ], 
		       [comment(['Calculate the face of the simplex: ', 
		                 'find the best point and the two worst points'])]
		      ),

        /*
         * try a reflection first
         *
         */
	synth_simplex_try(Offset_vector, Simplex, Worst1, ValuesW1, Center, Size, 
			  Try, Try_val,
			  -1, SFormula, 
			  'Reflect the simplex from the worst point (F = -1)',
			  Reflection),

        /*
         * try a reflection and expansion
         *
         */
	synth_simplex_try(Offset_vector, Simplex, Worst1, ValuesW1, Center, Size, 
			  Try, Try_val,
			  2, SFormula, 
			  'Reflection gives a better result, so try an additional extrapolation (F = 2)',
			  Extrapolation),

        /*
         * try a one-dimensional contraction
         *
         */
	synth_simplex_try(Offset_vector, Simplex, Worst1, ValuesW1, Center, Size, 
			  Try, Try_val,
			  1/2, SFormula, 
			  'Reflection gives a worse result, so try a one-dimensional contraction (F = 0.5)',
			  Contraction1),

        /*
         * try a multi-dimensional contraction
         *
         */
	pv_fresh_existential(K),
	index_make([K, dim(Ar_base, Size0)], Index_k),
	Simplex_ik =.. [Simplex, I, K],
	Simplex_bk =.. [Simplex, Best, K],
	ContractionN_eval = series([for([Index_k],
					assign(Simplex_ik, 
					       (Simplex_ik+Simplex_bk)/2, 
					       []),
					[]),
				    assign(Value_i, SFormula_i, 
			                   [comment('Re-evaluate expression')])
				   ],
				   []),
	ContractionN_loop = for([Index_i], 
				if(not(I = Best),
				   ContractionN_eval,
				   skip,
				   []
				  ), 
				[comment('Contract around the best point')]),

	ContractionN = if(Try_val =< Save_val,
			  ContractionN_loop,
			  skip,
			  []),

	Contraction  = if(Try_val =< ValuesW2,
			  series([assign(Save_val, ValuesW1,[]), 
				  Contraction1, 
				  ContractionN], 
				 []),
			  skip,
			  []
			 ),

        /*
         * now dispatch depending on the result of the reflection
         *
         */
	Dispatch = if(Try_val >= ValuesB,
		      Extrapolation,
		      Contraction,
		      []
		     ),

        /*
         * synthesize the convergence loop from the pieces
         *
         */
	Loop = while(converging([scalar(ValuesB), 	
				 scalar(ValuesW2), 	
				 scalar(ValuesW1)]),
		     series([Centroid, Sweep, Reflection, Dispatch], []),
		     [comment('Convergence loop')]
		    ),

        /*
         * synthesize the postamble assignments
         *
         */
	Simplex_best =.. [Simplex, Best],
	apack_unpack(Simplex_best, 
		     [comment('Copy best values back')], 
		     Offset_vector_copy, 
		     Post_assigns
		    ),

        %
        % Construct explanation
        %
	lex_enumerate_vars(Theta, XPTheta),
        XP = [
          'The function', expr(Formula), 'is then maximized w.r.t. the ',
          'goal ', XPTheta, ' using the simplex method. ' | XPSexp
        ],


	%
	% Compose program from pieces
	%
	Init      = series([CInit, Init_simplex, Init_values], 
			   [comment('Initialization')]),
        Program   = block(local([Simplex_decl, Try_decl, Values_decl, 
			         Try_val_decl, Save_val_decl, Center_decl,
			         Best_decl, Worst2_decl, Worst1_decl | CDecls]),
			  series([Init, Loop, Post_assigns], []),
                          [label(Simplex_label), comment(XP)]),

	assert_trace(trace_schema_inout,
		     'synth/schemas/numeric/simplex.pl', 
		     'Done Nelder-Mead simplex method').

/*************************************************************************
 *
 * Auxiliary predicates
 *
 *************************************************************************/

synth_simplex_init(Name, Size, [Dim, _], Offset_vector, Origin, Units, Init) :-
	index_make([I, Dim], Index),
	Simplex =.. [Name, I],
	apack_update(assign, Simplex, Offset_vector, Origin, Assigns),
	Init_origin = for([Index], 
			  Assigns,
			  [label(XPL1), comment('Copy origin into each point')]
			 ),
	list_zip(synth_simplex_init_units(Name), Offset_vector, Units, Incrs),
	Init_units = series(Incrs, 
			    [label(XPL2), comment('Add variances in each dimension')]
			   ),
	Init = series([Init_origin, Init_units], [comment(XP)]),
	XP = [
	  'Initialization of the simplex structure ', Name, ';', pp_nl,
	  'the simplex consists of (', Size, '+1) points in the the ',
	  Size, '-dimensional space. The first point (or origin) is ',
	  'constructed from an estimate of each variable\'s mean ',
	  labelref(XPL1), '. The remaining ', Size, ' points are each ',
	  'constructed by adding estimates of the variables\' variances ',
	  'in each dimension ', labelref(XPL2), '.'
	].

synth_simplex_init_units(Name, [Offset, Pair], Stmt) :-
	Offset = [_, Index, _],
	LHS    =.. [Name, 1 + Index],
	apack_update_(assign_plus, LHS, [Offset, Pair], Stmt).


synth_simplex_try(Offset_vector, Simplex, Worst, Worst_val, Center, Size, 
		    Try, Try_val, Factor, Formula, 
		    XP, Program) :-

	Ar_base   = 0,
	Ar_offset = 1,

	pv_fresh_existential(I),
	index_make([I, dim(Ar_base, Size-Ar_offset)], Index_i),

	Center_i  =.. [Center, I],
	Simplex_i =.. [Simplex, Worst, I],
	Try_i     =.. [Try, I],
	Factor1   = (1 - Factor) / Size,
	Factor2   = Factor1 - Factor,
	apack_substitute(Try, Offset_vector, Formula, Formula_try),

	Replace_loop = for([Index_i],
			   series([assign_plus(Center_i, Try_i-Simplex_i, []),
				   assign(Simplex_i, Try_i, [])
				  ], []),
			   []
			  ),
	Replace = if(Try_val > Worst_val,
		     series([assign(Worst_val, Try_val, []), Replace_loop],
			    []
			   ),
		     skip,
		     [comment('Replace worst point by trial point')]
		    ),
	Program = series([for([Index_i],
			      assign(Try_i, Center_i*Factor1-Simplex_i*Factor2, 
				     []),
			      [comment('Calculate trial point coordinates')]),
			  assign(Try_val, Formula_try, 
				 [comment('Evaluate expression at trial point ')]), 
			  Replace],
			 [comment(XP)]
			).




/*
 * synth_term_redexes/3 - parameter permutation
 *   
 */
synth_term_redexes(Term, Pattern, Redexes) :-
	term_redexes(Pattern, Term, Redexes).
